import pandas as pd
import numpy as np
import os, sys
import uuid
import traceback
import argparse
from datetime import timedelta
import math
import time
import warnings
warnings.filterwarnings("ignore")

# Append maxpermissible values folder
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

# Append scripts folder
parentdir = os.path.dirname(parentdir)
sys.path.append(parentdir)

from lib.Thresholds import Thresholds, METHODS_CLASS_DICT
import json
from intelli_utilities import BaiException, LOCAL_TIME_ZONE
from metrics.calculate_metrics import calculate_metrics
import multiprocessing
from collections import deque 
from metrics.metric_utils import *
from filelock import FileLock
from datetime import datetime

import logging
logging.basicConfig(filename='log/jobs.log', encoding='utf-8', level=logging.ERROR)

# Where it is executed in cluster or locally
EXECUTION_CONTEXT = None
SPOOL_DIR_UUID = ''

def compute_result(train_data_df:pd.DataFrame, test_data_df:pd.DataFrame, job_config:dict, thresholds:Thresholds, queue, config:dict):
    dependent_var_colname = get_node_unique_identifier(config['dependent_var']['pmModel'])
    threshold_instance = config['threshold_class'](dependent_variables=[dependent_var_colname], explanatory_vars=config['explanatory_var'], use_stats=config['stats'], model=config['model_class'])

    # # Set dependent var as value named col. since classes and outlier detection looking vor 'value'
    # train_data_df['value'] = train_data_df['value']
    # if test_data_df.empty == False:
    #     test_data_df['value'] = test_data_df['value']
    
    result = None
    try:          
        if job_config['trainMethod'] == 'isolatedWindows':                
            results_df = thresholds.train_static_data(threshold_instance, train_data_df, test_data_df)

        if job_config['trainMethod'] == 'movingWindows':
            results_df, _, _, _ = thresholds.train_rolling_data(threshold_instance, train_data_df, job_config['periodsDay']['train'], job_config['periodsDay']['retrain'], True)


        if not results_df.empty:
            mseEScaled = calculate_metrics(results_df, dependent_var_colname)
            # mseEScaled can be very high due to binning error in #45108
            # mseEScaled can be 0.0 due to most Y_pred = nan. Probably a consequent of to many missing infos (minBinElements)
            if mseEScaled > 10000 or mseEScaled != 0.0:
                missingDataCount = 'N/A'
                if hasattr(threshold_instance, 'missing_data'):
                    missingDataCount = len(threshold_instance.missing_data)
                if np.isnan(mseEScaled) == False:
                    result = {
                        'uuid': str(uuid.uuid4()),
                        'config': {
                            'dependentVar': config['dependent_var'],
                            'task': {
                                'name': 'maxpermissiblevalues',
                                'method': config['threshold_name'],
                                'model': config['model_name'],
                            },
                            'explanatoryVariables': config['explanatory_var_id'],
                            'missingDataCount': missingDataCount,
                            'metric': round(mseEScaled, 3),
                            'stats': config['stats'],
                            'tuningJobUuid': '',
                        }  
                    }
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        error = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        logging.error(error)

    # put result to compute remaining percentages
    queue.put(result)
        
    return result          
    
def send_progress(progress:dict):
    if EXECUTION_CONTEXT == 'local':
        print(json.dumps(progress), flush=True)
    else:
        lock = FileLock(f"{SPOOL_DIR_UUID}/progress.lock")
        lock.acquire()
        try:
            with open(f"{SPOOL_DIR_UUID}/progress.json", 'w') as progress_file:
                json.dump(progress, progress_file)   
        finally:
            # TODO: Check if we can use either release or remove
            lock.release()
            os.remove(f"{SPOOL_DIR_UUID}/progress.lock") # needed?
     
            
if __name__ == "__main__":
    
    start = time.time()
    parser = argparse.ArgumentParser()
    parser.add_argument('--jobUuid', help='Uuid of the page to load')
    parser.add_argument('--context', help='execution target. ie. local, cluster')
    parser.add_argument('--token', help='Security token for services')
    parser.add_argument('--processManager', help='Base url of the process manager')
    args = parser.parse_args()

    try:
        JOB_UUID = args.jobUuid
        EXECUTION_CONTEXT = args.context        

        SPOOL_DIR_UUID = f"/opt/intelli-service/spool/{JOB_UUID}"        
        if EXECUTION_CONTEXT == 'local':
            SPOOL_DIR_UUID =  f"spool/{JOB_UUID}"
        
        input_file = open(f"{SPOOL_DIR_UUID}/input.json")
        input_data = json.load(input_file)
        input_file.close()
        job = input_data['job']
        job_config = job['config']

        data_df = pd.DataFrame(input_data['dataFrame'])
        data_df = data_df.set_index('dateTime', drop=False)
        data_df.index = pd.to_datetime(data_df.index)                

        dataframes = []
        for date in job_config['trainingDates']:
            dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
            dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)
            dataframes.append(data_df[dt_start:dt_end])
        train_data_df = pd.concat(dataframes)

        test_data_df = pd.DataFrame()
        if len(job_config['measureDates']) > 0:
            dataframes = []
            for date in job_config['measureDates']:
                dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
                dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)
                dataframes.append(data_df[dt_start:dt_end])
            test_data_df = pd.concat(dataframes)

        # Fetch and combine data
        thresholds = Thresholds()
        interval = job_config['interval']        

        response = {
            'jobUuid': job['uuid'], 
            'status': 'computing',
            'percentage': 0, 
            'configListLength': 0, 
            'computedConfigs': 0,
        }       
              
        cpus = job['cpu']
        pool = multiprocessing.Pool(processes=cpus)
        m = multiprocessing.Manager()
        queue = m.Queue()                
                            
        temporal_var_name_list, explanatory_temporal_vars = get_temporal_var_list(interval)
                          
        # Synchronizes mandatory flag between important features sets and nodes
        important_feature_sets = set_mandatory_flag(job_config['importantFeaturesSets'], job_config['nodes'])                
                                                         
        #----- Create combinations
        combinations = get_dependent_independent_vars_combinations(important_feature_sets)
        # ----- Create config list
        config_list = []
        for combination in combinations:
            for independent_vars in combination['independent_vars']:
                explanatory_var_name_list = get_colname_list(independent_vars)              
                
                for threshold_dict in METHODS_CLASS_DICT.items():
                    if threshold_dict[0] == 'thresholdsGrouping':  
                        config_list.append((
                            train_data_df, 
                            test_data_df, 
                            job_config, 
                            thresholds, 
                            queue,
                            {
                                'dependent_var': combination['dependent_var'],
                                'threshold_name': threshold_dict[0],
                                'threshold_class': threshold_dict[1],
                                'model_name': '',
                                'model_class': '',                                
                                'explanatory_var': explanatory_var_name_list,
                                'explanatory_var_id': independent_vars,                        
                                'stats': 'mean', # for stats in ['median', 'mean']: // move to hyperparameter tuning   
                            }
                        ))     

                    if threshold_dict[0] == 'thresholdsRegression':
                            for model in thresholds._regressor_model_list.items():
                                config_list.append((
                                    train_data_df, 
                                    test_data_df, 
                                    job_config, 
                                    thresholds, 
                                    queue,
                                    {
                                        'dependent_var': combination['dependent_var'],
                                        'threshold_name': threshold_dict[0],
                                        'threshold_class': threshold_dict[1],
                                        'model_name': model[0],
                                        'model_class': model[1](),
                                        'explanatory_var': explanatory_var_name_list,
                                        'explanatory_var_id': independent_vars,
                                        'stats': '',                                        
                                    }
                                ))

        response['configListLength'] = len(config_list)
        send_progress(response)

        #------- Calculate metrics         
        chunk_size = len(config_list) // job['cpu']                    
        results = pool.starmap_async(compute_result, config_list, chunksize=chunk_size)

        # monitor task results
        while True:
            if results.ready():
                break
            else:
                computed_tasks = queue.qsize()
                response['computedConfigs'] = computed_tasks
                response['percentage'] = round((computed_tasks / len(config_list)) * 100)                        
                send_progress(response)
                results.wait(1)                
        pool.close()
        pool.join()
        
        #---- Split dependent variables with their metrics and reduce the amount of configs by max_configs_per_dependent_variable
        max_configs_per_dependent_variable = 50
        dependent_vars_results = {}
        for result in results.get():
            # If error occurs in task computation, result will equal to None.            
            if result != None:
                dependent_var_uuid = result['config']['dependentVar']['uuid']
                if dependent_var_uuid not in dependent_vars_results:
                    dependent_vars_results[dependent_var_uuid] = deque([], maxlen=max_configs_per_dependent_variable)
                
                dependent_vars_result_configs = dependent_vars_results[dependent_var_uuid]
                if len(dependent_vars_result_configs) > 0:     
                    dependent_vars_result_configs = deque(sorted(dependent_vars_result_configs, key=lambda x: x['config']['metric'], reverse=False), maxlen=max_configs_per_dependent_variable)
                    if len(dependent_vars_result_configs) < max_configs_per_dependent_variable:
                        dependent_vars_result_configs.appendleft(result)
                    elif dependent_vars_result_configs[-1]['config']['metric'] > result['config']['metric']:
                        dependent_vars_result_configs.appendleft(result)
                else:
                    dependent_vars_result_configs.appendleft(result)
                dependent_vars_results[dependent_var_uuid] = dependent_vars_result_configs
                
        results = []                                                      
        for result_config in dependent_vars_results.items():
            results = results + list(result_config[1])
     
        #--- Sorting to have best metric at very top
        results.sort(key=lambda x: x['config']['metric'], reverse=False)

        with open(f"{SPOOL_DIR_UUID}/output.json", 'w') as outfile:
            json.dump(results, outfile)            

        end = time.time()        
        response['computedConfigs'] = len(config_list)
        response['totalDuration'] = str(timedelta(seconds=math.ceil(end - start))) 
        response['status'] = 'done'
        response['percentage'] = 100
        send_progress(response)
    except BaiException as be:
        error = "{}".format(be)
        logging.error(error)
        response = {
            'errorMsg': error,
        }
        send_progress(response)
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        error = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        logging.error(error)        
        response = {
            'errorMsg': error,
        }
        send_progress(response)
